//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      Device/Data/Powerfield.h
//! @brief     Defines and implements templated class Powerfield.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#ifndef BORNAGAIN_DEVICE_DATA_POWERFIELD_H
#define BORNAGAIN_DEVICE_DATA_POWERFIELD_H

#include "Base/Axis/Bin.h"
#include "Base/Axis/FixedBinAxis.h"
#include "Base/Py/PyObject.h"
#include "Base/Types/OwningVector.h"
#include "Base/Util/Assert.h"
#include "Base/Util/ThreadInfo.h"
#include "Device/Data/LLData.h"
#include "Device/Data/PowerfieldIterator.h"
#include <sstream>

using std::size_t;

//! Templated class to store data in multi-dimensional space.
//! Used with type T = double, CumulativeValue, bool
//! @ingroup tools

class IField {
public:
    //! Returns number of dimensions.
    size_t rank() const { return m_axes.size(); }

protected:
    //! Returns serial number of axis with given name
    size_t getAxisIndex(const std::string& axis_name) const;

    //! Checks if given axis name exists
    bool axisNameExists(const std::string& axis_name) const;

    OwningVector<IAxis> m_axes;
};

template <class T>
class Powerfield : public IField {
public:
    using value_type = T;

    Powerfield();
    Powerfield(const Powerfield&) = delete;
    Powerfield(Powerfield&&);
    const Powerfield& operator=(const Powerfield&) = delete;
    ~Powerfield();
    Powerfield* clone() const;

    void copyFrom(const Powerfield<T>& other);
    template <class U>
    void copyShapeFrom(const Powerfield<U>& other);
    Powerfield<double>* meanValues() const;

    void addAxis(const IAxis& new_axis);
    void addAxis(const std::string& name, size_t size, double start, double end);

    //! Returns axis with given serial number
    const IAxis& axis(size_t serial_number) const;

    // retrieve basic info

    //! Returns total size of data buffer (product of bin number in every dimension).
    size_t getAllocatedSize() const
    {
        if (m_ll_data)
            return m_ll_data->getTotalSize();
        return 0;
    }

    //! Returns all sizes of its axes
    std::vector<size_t> getAllSizes() const;

    //! Returns copy of raw data vector
    std::vector<T> getRawDataVector() const;

    //! Returns sum of all values in the data structure
    T totalSum() const;

    //! Returns maximum value in the data structure
    T max() const;

    // iterators

    friend class PowerfieldIterator<T, Powerfield<T>>;
    friend class PowerfieldIterator<const T, const Powerfield<T>>;

    //! Read/write iterator type
    using iterator = PowerfieldIterator<T, Powerfield<T>>;

    //! Read-only iterator type
    using const_iterator = PowerfieldIterator<const T, const Powerfield<T>>;

    //! Returns  read/write iterator that points to the first element
    iterator begin();

    //! Returns  read-only iterator that points to the first element
    const_iterator begin() const;

    //! Returns  read/write iterator that points to the one past last element
    iterator end() { return iterator(this, getAllocatedSize()); }

    //! Returns  read-only iterator that points to the one past last element
    const_iterator end() const { return const_iterator(this, getAllocatedSize()); }

    // coordinate and index functions

    //! Returns vector of axes indices for given global index
    //! @param global_index The global index of this data structure.
    //! @return Vector of bin indices for all axes defined
    std::vector<int> getAxesBinIndices(size_t global_index) const;

    //! Returns axis bin index for given global index
    //! @param global_index The global index of this data structure.
    //! @param i_selected_axis Serial number of selected axis.
    //! @return Corresponding bin index for selected axis
    size_t getAxisBinIndex(size_t global_index, size_t i_selected_axis) const;

    //! Returns axis bin index for given global index
    //! @param global_index The global index of this data structure.
    //! @param axis_name The name of selected axis.
    //! @return Corresponding bin index for selected axis
    size_t getAxisBinIndex(size_t global_index, const std::string& axis_name) const;

    //! Returns global index for specified indices of axes
    //! @param axes_indices Vector of axes indices for all specified axes in this dataset
    //! @return Corresponding global index
    size_t toGlobalIndex(const std::vector<unsigned>& axes_indices) const;

    //! Returns global index for specified axes values
    //! @param coordinates Vector of axes coordinates for all specified axes in this dataset
    //! @return Closest global index
    size_t findGlobalIndex(const std::vector<double>& coordinates) const;

    //! Returns the value of selected axis for given global_index.
    //! @param global_index The global index of this data structure.
    //! @param i_selected_axis Serial number of selected axis.
    //! @return corresponding bin center of selected axis
    double getAxisValue(size_t global_index, size_t i_selected_axis) const;

    //! Returns the value of selected axis for given global_index.
    //! @param global_index The global index of this data structure.
    //! @param axis_name The name of selected axis.
    //! @return corresponding bin center of selected axis
    double getAxisValue(size_t global_index, const std::string& axis_name) const;

    //! Returns values on all defined axes for given globalbin number
    //! @param global_index The global index of this data structure.
    //! @return Vector of corresponding bin centers
    std::vector<double> getAxesValues(size_t global_index) const;

    //! Returns bin of selected axis for given global_index.
    //! @param global_index The global index of this data structure.
    //! @param i_selected_axis Serial number of selected axis.
    //! @return Corresponding Bin1D object
    Bin1D getAxisBin(size_t global_index, size_t i_selected_axis) const;

    //! Returns bin of selected axis for given global_index.
    //! @param global_index The global index of this data structure.
    //! @param axis_name The name of selected axis.
    //! @return Corresponding Bin1D object
    Bin1D getAxisBin(size_t global_index, const std::string& axis_name) const;

    // modifiers

    //! Sets object into initial state (no dimensions, data)
    void clear();

    //! Sets content of output data to specific value
    void setAllTo(const T& value);

    //! Adds 'rank' axes with indicated sizes
    void setAxisSizes(size_t rank, int* n_dims);

    //! Sets new values to raw data vector
    void setRawDataVector(const std::vector<T>& data_vector);

    //! Sets new values to raw data array
    void setRawDataArray(const T* source);

    //! addition-assignment operator for two output data
    const Powerfield<T>& operator+=(const Powerfield<T>& right);

    //! substraction-assignment operator for two output data
    const Powerfield<T>& operator-=(const Powerfield<T>& right);

    //! division-assignment operator for two output data
    const Powerfield<T>& operator/=(const Powerfield<T>& right);

    //! multiplication-assignment operator for two output data
    const Powerfield<T>& operator*=(const Powerfield<T>& right);

    //! Returns value or summed value, depending on T
    double getValue(size_t index) const;

    //! indexed accessor
    T& operator[](size_t index)
    {
        ASSERT(m_ll_data);
        return (*m_ll_data)[index];
    }

    //! indexed accessor (const)
    const T& operator[](size_t index) const
    {
        ASSERT(m_ll_data);
        return (*m_ll_data)[index];
    }

    // helpers

    //! Returns true if object have same dimensions and number of axes bins
    template <class U>
    bool hasSameDimensions(const Powerfield<U>& right) const;

    //! Returns true if objects a) have same dimensions b) bin boundaries of axes coincide
    template <class U>
    bool hasSameShape(const Powerfield<U>& right) const;

#ifdef BORNAGAIN_PYTHON
    //! Returns data as Python numpy array
    PyObject* getArray() const;
#endif

    //! Returns true if object is correctly initialized
    bool isInitialized() const;

    //! Allocates memory for current dimensions configuration
    void allocate();

    const Powerfield<double>& normalizedToMaximum();

private:
    LLData<T>* m_ll_data;
};

#ifndef USER_API

// --------------------------------------------------------------------- //
// Implementation
// --------------------------------------------------------------------- //

template <class T>
Powerfield<T>::Powerfield(Powerfield<T>&& other)
    : IField(other)
{
    m_ll_data = other.m_ll_data;

    other.m_ll_data = nullptr;
    other.allocate();
}

template <class T>
Powerfield<T>::Powerfield()
    : m_ll_data(nullptr)
{
    allocate();
}

template <class T>
Powerfield<T>::~Powerfield()
{
    clear();
    delete m_ll_data;
}

template <class T>
Powerfield<T>* Powerfield<T>::clone() const
{
    auto* result = new Powerfield<T>();
    result->m_axes = m_axes;
    (*result->m_ll_data) = *m_ll_data;
    return result;
}

template <class T>
void Powerfield<T>::copyFrom(const Powerfield<T>& other)
{
    clear();
    m_axes = other.m_axes;
    delete m_ll_data;
    m_ll_data = 0;
    if (other.m_ll_data)
        m_ll_data = new LLData<T>(*other.m_ll_data);
}

template <class T>
template <class U>
void Powerfield<T>::copyShapeFrom(const Powerfield<U>& other)
{
    clear();
    size_t rank = other.rank();
    for (size_t i = 0; i < rank; ++i)
        addAxis(other.axis(i));
}

template <class T>
Powerfield<double>* Powerfield<T>::meanValues() const
{
    auto* result = new Powerfield<double>();
    result->copyShapeFrom(*this);
    result->allocate();
    for (size_t i = 0; i < m_ll_data->getTotalSize(); ++i)
        (*result)[i] = getValue(i);
    return result;
}

template <class T>
void Powerfield<T>::addAxis(const IAxis& new_axis)
{
    ASSERT(!axisNameExists(new_axis.axisName()));
    if (new_axis.size() > 0) {
        m_axes.emplace_back(new_axis.clone());
        allocate();
    }
}

template <class T>
void Powerfield<T>::addAxis(const std::string& name, size_t size, double start, double end)
{
    ASSERT(!axisNameExists(name));
    FixedBinAxis new_axis(name, size, start, end);
    addAxis(new_axis);
}

template <class T>
const IAxis& Powerfield<T>::axis(size_t serial_number) const
{
    return *m_axes[serial_number];
}

template <class T>
inline std::vector<size_t> Powerfield<T>::getAllSizes() const
{
    ASSERT(m_ll_data);
    std::vector<size_t> result;
    for (size_t i = 0; i < rank(); ++i) {
        int dim = m_ll_data->dimensions()[i];
        result.push_back(dim);
    }
    return result;
}

template <class T>
inline std::vector<T> Powerfield<T>::getRawDataVector() const
{
    ASSERT(m_ll_data);
    std::vector<T> result;
    for (size_t i = 0; i < getAllocatedSize(); ++i)
        result.push_back((*m_ll_data)[i]);
    return result;
}

template <class T>
typename Powerfield<T>::iterator Powerfield<T>::begin()
{
    typename Powerfield<T>::iterator result(this);
    return result;
}

template <class T>
typename Powerfield<T>::const_iterator Powerfield<T>::begin() const
{
    typename Powerfield<T>::const_iterator result(this);
    return result;
}

template <class T>
std::vector<int> Powerfield<T>::getAxesBinIndices(size_t global_index) const
{
    ASSERT(m_ll_data);
    size_t remainder = global_index;
    std::vector<int> result;
    result.resize(m_ll_data->rank());
    for (size_t i = 0; i < m_ll_data->rank(); ++i) {
        result[m_ll_data->rank() - 1 - i] =
            (int)(remainder % m_axes[m_ll_data->rank() - 1 - i]->size());
        remainder /= m_axes[m_ll_data->rank() - 1 - i]->size();
    }
    return result;
}

template <class T>
size_t Powerfield<T>::getAxisBinIndex(size_t global_index, size_t i_selected_axis) const
{
    ASSERT(m_ll_data);
    size_t remainder(global_index);
    for (size_t i = 0; i < m_ll_data->rank(); ++i) {
        size_t i_axis = m_ll_data->rank() - 1 - i;
        size_t result = remainder % m_axes[i_axis]->size();
        if (i_selected_axis == i_axis)
            return result;
        remainder /= m_axes[i_axis]->size();
    }
    ASSERT(0);
}

template <class T>
size_t Powerfield<T>::getAxisBinIndex(size_t global_index, const std::string& axis_name) const
{
    return getAxisBinIndex(global_index, getAxisIndex(axis_name));
}

template <class T>
size_t Powerfield<T>::toGlobalIndex(const std::vector<unsigned>& axes_indices) const
{
    ASSERT(m_ll_data);
    ASSERT(axes_indices.size() == m_ll_data->rank());
    size_t result = 0;
    size_t step_size = 1;
    for (size_t i = m_ll_data->rank(); i > 0; --i) {
        ASSERT(axes_indices[i - 1] < m_axes[i - 1]->size());
        result += axes_indices[i - 1] * step_size;
        step_size *= m_axes[i - 1]->size();
    }
    return result;
}

template <class T>
size_t Powerfield<T>::findGlobalIndex(const std::vector<double>& coordinates) const
{
    ASSERT(m_ll_data);
    ASSERT(coordinates.size() == m_ll_data->rank());
    std::vector<unsigned> axes_indexes;
    axes_indexes.resize(m_ll_data->rank());
    for (size_t i = 0; i < m_ll_data->rank(); ++i)
        axes_indexes[i] = static_cast<unsigned>(m_axes[i]->findClosestIndex(coordinates[i]));
    return toGlobalIndex(axes_indexes);
}

template <class T>
double Powerfield<T>::getAxisValue(size_t global_index, size_t i_selected_axis) const
{
    auto axis_index = getAxisBinIndex(global_index, i_selected_axis);
    return (*m_axes[i_selected_axis])[axis_index];
}

template <class T>
double Powerfield<T>::getAxisValue(size_t global_index, const std::string& axis_name) const
{
    return getAxisValue(global_index, getAxisIndex(axis_name));
}

template <class T>
std::vector<double> Powerfield<T>::getAxesValues(size_t global_index) const
{
    std::vector<int> indices = getAxesBinIndices(global_index);
    std::vector<double> result;
    for (size_t i_axis = 0; i_axis < indices.size(); ++i_axis)
        result.push_back((*m_axes[i_axis])[indices[i_axis]]);
    return result;
}

template <class T>
Bin1D Powerfield<T>::getAxisBin(size_t global_index, size_t i_selected_axis) const
{
    auto axis_index = getAxisBinIndex(global_index, i_selected_axis);
    return m_axes[i_selected_axis]->bin(axis_index);
}

template <class T>
Bin1D Powerfield<T>::getAxisBin(size_t global_index, const std::string& axis_name) const
{
    return getAxisBin(global_index, getAxisIndex(axis_name));
}

template <class T>
inline T Powerfield<T>::totalSum() const
{
    ASSERT(m_ll_data);
    return m_ll_data->getTotalSum();
}

template <class T>
inline T Powerfield<T>::max() const
{
    ASSERT(m_ll_data);
    return m_ll_data->max();
}

template <class T>
void Powerfield<T>::clear()
{
    m_axes.clear();
    allocate();
}

template <class T>
void Powerfield<T>::setAllTo(const T& value)
{
    ASSERT(m_ll_data);
    m_ll_data->setAll(value);
}

template <class T>
void Powerfield<T>::setAxisSizes(size_t rank, int* n_dims)
{
    clear();
    std::string basename("axis");
    for (size_t i = 0; i < rank; ++i) {
        std::ostringstream name;
        name << basename << i;
        addAxis(name.str(), n_dims[i], 0.0, (double)(n_dims[i] - 1));
    }
}

template <class T>
const Powerfield<T>& Powerfield<T>::operator+=(const Powerfield<T>& right)
{
    ASSERT(m_ll_data);
    *this->m_ll_data += *right.m_ll_data;
    return *this;
}

template <class T>
const Powerfield<T>& Powerfield<T>::operator-=(const Powerfield<T>& right)
{
    ASSERT(m_ll_data);
    *this->m_ll_data -= *right.m_ll_data;
    return *this;
}

template <class T>
const Powerfield<T>& Powerfield<T>::operator*=(const Powerfield<T>& right)
{
    ASSERT(m_ll_data);
    *this->m_ll_data *= *right.m_ll_data;
    return *this;
}

template <class T>
bool Powerfield<T>::isInitialized() const
{
    if (!m_ll_data)
        return false;
    if (rank() != m_ll_data->rank())
        return false;
    if (!rank())
        return false;
    return true;
}

template <class T>
const Powerfield<T>& Powerfield<T>::operator/=(const Powerfield<T>& right)
{
    ASSERT(m_ll_data);
    *this->m_ll_data /= *right.m_ll_data;
    return *this;
}

template <class T>
void Powerfield<T>::allocate()
{
    delete m_ll_data;
    size_t rank = m_axes.size();
    auto* dims = new int[rank];
    for (size_t i = 0; i < rank; ++i)
        dims[i] = (int)axis(i).size();
    m_ll_data = new LLData<T>(rank, dims);
    T default_value = {};
    m_ll_data->setAll(default_value);
    delete[] dims;
}

template <class T>
inline void Powerfield<T>::setRawDataVector(const std::vector<T>& data_vector)
{
    ASSERT(data_vector.size() == getAllocatedSize());
    for (size_t i = 0; i < getAllocatedSize(); ++i)
        (*m_ll_data)[i] = data_vector[i];
}

template <class T>
inline void Powerfield<T>::setRawDataArray(const T* source)
{
    for (size_t i = 0; i < getAllocatedSize(); ++i)
        (*m_ll_data)[i] = source[i];
}

//! Returns true if object have same dimensions
template <class T>
template <class U>
inline bool Powerfield<T>::hasSameDimensions(const Powerfield<U>& right) const
{
    if (!isInitialized())
        return false;
    if (!right.isInitialized())
        return false;
    if (rank() != right.rank())
        return false;
    for (size_t i_axis = 0; i_axis < rank(); ++i_axis)
        if (axis(i_axis).size() != right.axis(i_axis).size())
            return false;
    return true;
}

//! Returns true if object have same dimensions and shape of axis
template <class T>
template <class U>
bool Powerfield<T>::hasSameShape(const Powerfield<U>& right) const
{
    if (!hasSameDimensions(right))
        return false;

    for (size_t i = 0; i < m_axes.size(); ++i)
        if (axis(i) != right.axis(i))
            return false;
    return true;
}

#ifdef BORNAGAIN_PYTHON
//! Returns data as Python numpy array
template <>
PyObject* Powerfield<double>::getArray() const;
#endif

#endif // USER_API
#endif // BORNAGAIN_DEVICE_DATA_POWERFIELD_H
